/* A simple solver of linear equation systems
 * Author: Wojciech Komorowski (ot.elvag@xetroc, moc.liamtoh@flowagig)
 * Date: 2006-10-06
 * Version 1.0.0
 * License: GNU General Public License, version 2
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

void outOfMem(void);

int main(int argc, char** argv){

  int n;  /* size of eq. system */

  int i, j, k, max_row;
  long double tmp;

  long double** A = NULL;
  long double* X = NULL;

  if (argc > 1){
    /* determine size of eq.sys. by number
       of parameters passed as arguments
    */
    n = (-1 + sqrt(4*argc - 3))/2;
  }
  else{
    do{
      printf("Size of the system (0 to quit)?: ");
      if (scanf("%d", &n) == 0){
	printf("Size is an integer!\n");
	scanf("%*[^\n]"); /* skip to the end of the line */
	scanf("%*1[\n]"); /* skip one newline */
      } else break;
    } while(1);

    n = abs(n);
    if (!n) return 0;
  }

  A = (long double**)malloc(n*sizeof(long double*));
  X = (long double*)malloc(n*sizeof(long double));

  if (A == NULL || X == NULL){
    free(A);
    free(X);
    if (argc == 1) outOfMem();
    else exit(1);
  }

  /* initialize arrays
     A[n][n+1] for augmented matrix [A | x]
     X[n] for the solutions vector
  */
  for (i = 0; i < n; i++){
    A[i] = (long double*)malloc((n+1)*sizeof(long double));
    X[i] = 0;
    if (A[i] == NULL){
      free(A);
      free(X);
      if (argc == 1) outOfMem();
      else exit(1);
    }
  }

  if (argc == 1){
    for (i = 0; i < n; i++)
      for (j = 0; j < n+1; j++){
	printf("X%d,%d = ", i, j);
	if (scanf("%Lf", &tmp) == 0){  /* ask for variable coefficients */
	  printf("\nInvalid input!\n");
	  scanf("%*[^\n]"); /* skip to the end of the line */
	  scanf("%*1[\n]"); /* skip one newline */
	  j--;
	}
	else A[i][j] = tmp;  /* and save them */
      }

    /* print the matrix */
    printf("\nSystem matrix:\n");
    for (i = 0; i < n; i++)
      printf("X[%d]\t\t", i+1);
    printf("=\n");

    for (i = 0; i < n; i++)
      for (j = 0; j < n+1; j++){
	printf("%Lf\t", A[i][j]);
	if ((j+1) % (n+1) == 0)
	  printf("\n");
      }
  }
  else{
    k = 1;
    for (i = 0; i < n; i++)
      for (j = 0; j < n+1; j++){
	if (sscanf(argv[k], "%Lf", &tmp) == 0){
	  free(A);
	  free(X);
	  exit(2);
	}
	A[i][j] = tmp;
	k++;
      }
  }

  /* reduce matrix to row echelon form */
  for (i = 0; i < n; i++){
    max_row = i;
    for (j = i; j < n; j++)
      if (fabs(A[j][i]) > fabs(A[max_row][i]))
	max_row = j;

    for (j = 0; j < n+1; j++){
      tmp = A[i][j];
      A[i][j] = A[max_row][j];
      A[max_row][j] = tmp;
    }

    for (j = i+1; j < n; j++)
      for (k = n; k >= i; k--){
	if (A[i][i] == 0.0) continue;
	A[j][k] = A[j][k] - A[j][i]*A[i][k]/A[i][i];
      }
  }

  /* back substitute */
  for (i = n-1; i >= 0; i--){
    if (A[i][i] == 0.0){
      if (argc == 1){
	printf("\nSystem matrix does not have an inverse and thus is unsolvable in absolutes...\n");
	printf("Giving up!\n");
      }
      free(A);
      free(X);
      exit(1);
    }
    for (j = n-1; j > i; j--)
      X[i] = X[i] - A[i][j]*X[j]/A[i][i];

    X[i] = X[i] + A[i][n]/A[i][i];
  }

  if (argc == 1){
    /* display solutions matrix */
    for (i = 0; i < n; i++)
      printf("\nX[%d] = %Lf", i+1, X[i]);
    printf("\n");
  }
  else{
    for (i = 0; i < n; i++)
      printf("%Lf\n", X[i]);
  }

  free(A);
  free(X);
  return 0;
}

void outOfMem(void)
{
  printf("\nError: Out Of Memory\n");
  exit(1);
}
